@@ -1,5 +1,134 @@
 """
 Making DAGs
+
+In it's simplest form, consider this:
+
+>>> from meshed import DAG
+>>>
+>>> def this(a, b=1):
+...     return a + b
+...
+>>> def that(x, b=1):
+...     return x * b
+...
+>>> def combine(this, that):
+...     return (this, that)
+...
+>>>
+>>> dag = DAG((this, that, combine))
+>>> print(dag.synopsis_string())
+x,b -> that_ -> that
+a,b -> this_ -> this
+this,that -> combine_ -> combine
+
+But don't be fooled: There's much more to it!
+
+
+FAQ and Troubleshooting
+=======================
+
+DAGs and Pipelines
+------------------
+
+>>> from functools import partial
+>>> from meshed import DAG
+>>> def chunker(sequence, chk_size: int):
+...     return zip(*[iter(sequence)] * chk_size)
+>>>
+>>> my_chunker = partial(chunker, chk_size=3)
+>>>
+>>> vec = range(8)  # when appropriate, use easier to read sequences
+>>> list(my_chunker(vec))
+[(0, 1, 2), (3, 4, 5)]
+
+Oh, that's just a ``my_chunker -> list`` pipeline!
+A pipeline is a subset of DAG, so let me do this:
+
+>>> dag = DAG([my_chunker, list])
+>>> dag(vec)
+Traceback (most recent call last):
+...
+TypeError: missing a required argument: 'sequence'
+
+What happened here?
+You're assuming that saying ``[my_chunker, list]`` is enough for DAG to know that
+what you meant is for ``my_chunker`` to feed it's input to ``list``.
+Sure, DAG has enough information to do so, but the default connection policy doesn't
+assume that it's a pipeline you want to make.
+In fact, the order you specify the functions doesn't have an affect on the connections
+with the default connection policy.
+
+See what the signature of ``dag`` is:
+
+>>> from inspect import signature
+>>> str(signature(dag))
+'(iterable=(), /, sequence, *, chk_size: int = 3)'
+
+So dag actually works just fine. Here's the proof:
+
+>>> dag([1,2,3], vec)  # doctest: +SKIP
+([1, 2, 3], <zip object at 0x104d7f080>)
+
+It's just not what you might have intended.
+
+Your best bet to get what you intended is to be explicit.
+
+The way to be explicit is to not specify functions alone, but ``FuncNodes`` that
+wrap them, along with the specification
+the ``name`` the function will be referred to by,
+the names that it's parameters should ``bind`` to (that is, where the function
+will get it's import arguments from), and
+the ``out`` name of where it should be it's output.
+
+In the current case a fully specified DAG would look something like this:
+
+>>> from meshed import FuncNode
+>>> dag = DAG(
+...     [
+...         FuncNode(
+...             func=my_chunker,
+...             name='chunker',
+...             bind=dict(sequence='sequence', chk_size='chk_size'),
+...             out='chks'
+...         ),
+...         FuncNode(
+...             func=list,
+...             name='gather_chks_into_list',
+...             bind=dict(iterable='chks'),
+...             out='list_of_chks'
+...         ),
+...     ]
+... )
+>>> list(dag(vec))
+[(0, 1, 2), (3, 4, 5)]
+
+But really, if you didn't care about the names of things,
+all you need in this case was to make sure that the output of ``my_chunker`` was
+fed to ``list``, and therefore the following was sufficient:
+
+>>> dag = DAG([
+...     FuncNode(my_chunker, out='chks'),  # call the output of chunker "chks"
+...     FuncNode(list, bind=dict(iterable='chks'))  # source list input from "chks"
+... ])
+>>> list(dag(vec))
+[(0, 1, 2), (3, 4, 5)]
+
+Connection policies are very useful when you want to define ways for DAG to
+"just figure it out" for you.
+That is, you want to tell the machine to adapt to your thoughts, not vice versa.
+We support such technological expectations!
+The default connection policy is there to provide one such ways, but
+by all means, use another!
+
+Does this mean that connection policies are not for production code?
+Well, it depends. The Zen of Python (``import this``)
+states "explicit is better than implicit", and indeed it's often
+a good fallback rule.
+But defining components and the way they should be assembled can go a long way
+in achieving consistency, separation of concerns, adaptability, and flexibility.
+All quite useful things. Also in production. Especially in production.
+That said it is your responsiblity to use the right policy for your particular context.
+
 """
 
 from contextlib import suppress
@@ -53,6 +182,10 @@ def find_first_free_name(prefix, exclude_names=(), start_at=2):
             i += 1
 
 
+class NameValidationError(ValueError):
+    """Use to indicate that there's a problem with a name or generating a valid name"""
+
+
 def mk_func_name(func, exclude_names=()):
     name = getattr(func, '__name__', '')
     if name == '<lambda>':
@@ -60,6 +193,8 @@ def mk_func_name(func, exclude_names=()):
     elif name == '':
         if isinstance(func, partial):
             return mk_func_name(func.func, exclude_names)
+        else:
+            raise NameValidationError(f"Can't make a name for func: {func}")
     return find_first_free_name(name, exclude_names)
 
 
@@ -87,7 +222,7 @@ def named_partial(func, *args, __name__=None, **keywords):
     >>> f = named_partial(print, sep='\\n')
     >>> f.__name__
     'print'
-    
+
     >>> f = named_partial(print, sep='\\n', __name__='now_partial_has_a_name')
     >>> f.__name__
     'now_partial_has_a_name'
@@ -97,7 +232,7 @@ def named_partial(func, *args, __name__=None, **keywords):
     return f
 
 
-from i2.deco import ch_func_to_all_pk
+from i2.signatures import ch_func_to_all_pk
 
 
 def hook_up(func, variables: MutableMapping, output_name=None):
@@ -231,8 +366,9 @@ def _mapped_extraction(src: dict, to_extract: dict):
 
 def underscore_func_node_names_maker(func: Callable, name=None, out=None):
     """This name maker will resolve names in the following fashion:
-     (1) look at the (func) name and out given as arguments, if None...
-     (3) use mk_func_name(func) to make names.
+
+     #. look at the (func) name and out given as arguments, if None...
+     #. use mk_func_name(func) to make names.
 
     It will use the mk_func_name(func)  itself for out, but suffix the same with
     an underscore to provide a mk_func_name.
@@ -244,7 +380,15 @@ def underscore_func_node_names_maker(func: Callable, name=None, out=None):
     if name is not None and out is not None:
         return name, out
 
-    name_of_func = mk_func_name(func)
+    try:
+        name_of_func = mk_func_name(func)
+    except NameValidationError as err:
+        err_msg = err.args[0]
+        err_msg += (
+            f'\nSuggestion: You might want to specify a name explicitly in '
+            f'FuncNode(func, name=name) instead of just giving me the func as is.'
+        )
+        raise NameValidationError(err_msg)
     if name is None and out is None:
         return name_of_func + '_', name_of_func
     elif out is None:
@@ -261,9 +405,63 @@ def duplicates(elements: Union[Iterable, Sized]):
         return []
 
 
+def _keys_and_values_are_strings_validation(d: dict):
+    for k, v in d.items():
+        if not isinstance(k, str):
+            raise ValidationError(f'Should be a str: {k}')
+        if not isinstance(v, str):
+            raise ValidationError(f'Should be a str: {v}')
+
+
+def _func_node_args_validation(func: Callable, name: str, bind: dict, out: str):
+    """Validates the four first arguments that are used to make a ``FuncNode``.
+    Namely, if not ``None``,
+
+    * ``func`` should be a callable
+
+    * ``name`` and ``out`` should be ``str``
+
+    * ``bind`` should be a ``Dict[str, str]``
+
+    """
+    if func is not None and not isinstance(func, Callable):
+        raise ValidationError(f'Should be callable: {func}')
+    if name is not None and not isinstance(name, str):
+        raise ValidationError(f'Should be a str: {name}')
+    if bind is not None:
+        if not isinstance(bind, dict):
+            raise ValidationError(f'Should be a dict: {bind}')
+        _keys_and_values_are_strings_validation(bind)
+    if out is not None and not isinstance(out, str):
+        raise ValidationError(f'Should be a str: {out}')
+
+
 def basic_node_validator(func_node):
+    """Validates a func node. Raises ValidationError if something wrong. Returns None.
+
+    Validates:
+
+    * that the ``func_node`` params are valid, that is, if not ``None``
+        * ``func`` should be a callable
+        * ``name`` and ``out`` should be ``str``
+        * ``bind`` should be a ``Dict[str, str]``
+    * that the names (``.name``, ``.out`` and all ``.bind.values()``)
+        * are valid python identifiers (alphanumeric or underscore not starting with
+          digit)
+        * are not repeated (no duplicates)
+    * that ``.bind.keys()`` are indeed present as params of ``.func``
+
+    """
+    _func_node_args_validation(
+        func_node.func, func_node.name, func_node.bind, func_node.out
+    )
     names = [func_node.name, func_node.out, *func_node.bind.values()]
 
+    names_that_are_not_strings = [name for name in names if not isinstance(name, str)]
+    if names_that_are_not_strings:
+        names_that_are_not_strings = ', '.join(map(str, names_that_are_not_strings))
+        raise ValidationError(f'Should be strings: {names_that_are_not_strings}')
+
     # Make sure there's no name duplicates
     _duplicates = duplicates(names)
     if _duplicates:
@@ -393,6 +591,7 @@ class FuncNode:
     node_validator: Callable = basic_node_validator
 
     def __post_init__(self):
+        _func_node_args_validation(self.func, self.name, self.bind, self.out)
         self.name, self.out = self.names_maker(self.func, self.name, self.out)
 
         # The wrapped function's signature will be useful
@@ -419,9 +618,9 @@ class FuncNode:
         an actual interface method. Additional control/constraints on read and writes
         can be implemented by providing a custom scope for that."""
         relevant_kwargs = dict(self.extractor(scope))
-        # print(scope, relevant_kwargs)
+        args, kwargs = self.sig.args_and_kwargs_from_kwargs(relevant_kwargs)
         output = call_somewhat_forgivingly(
-            self.func, (), relevant_kwargs, enforce_sig=self.sig
+            self.func, args, kwargs, enforce_sig=self.sig
         )
         if self.write_output_into_scope:
             scope[self.out] = output
@@ -446,7 +645,7 @@ class FuncNode:
 
     @classmethod
     def has_as_instance(cls, obj):
-        """Verify if obj is an instance of a FuncNode (or specific sub-class).
+        """Verify if ``obj`` is an instance of a FuncNode (or specific sub-class).
 
         The usefulness of this method is to not have to make a lambda with isinstance
         when filtering.
@@ -462,9 +661,10 @@ class FuncNode:
 def validate_that_func_node_names_are_sane(func_nodes: Iterable[FuncNode]):
     """Assert that the names of func_nodes are sane.
     That is:
-        - are valid dot (graphviz) names (we'll use str.isidentifier because lazy)
-        - All the func.name and func.out are unique
-        - more to come
+
+    * are valid dot (graphviz) names (we'll use str.isidentifier because lazy)
+    * All the ``func.name`` and ``func.out`` are unique
+    * more to come (TODO)...
     """
     func_nodes = list(func_nodes)
     node_names = [x.name for x in func_nodes]
@@ -525,12 +725,32 @@ def is_not_func_node(obj) -> bool:
     return not FuncNode.has_as_instance(obj)
 
 
-def extract_values(d: dict, keys: Iterable):
+def _extract_values(d: dict, keys: Iterable):
     """generator of values extracted from d for keys"""
     for k in keys:
         yield d[k]
 
 
+def extract_values(d: dict, keys: Iterable):
+    """Extract values from dict ``d``, returning them:
+
+    - as a tuple if len(keys) > 1
+
+    - a single value if len(keys) == 1
+
+    - None if not
+
+    This is used as the default extractor in DAG
+    """
+    tup = tuple(_extract_values(d, keys))
+    if len(tup) > 1:
+        return tup
+    elif len(tup) == 1:
+        return tup[0]
+    else:
+        return None
+
+
 def extract_items(d: dict, keys: Iterable):
     """generator of (k, v) pairs extracted from d for keys"""
     for k in keys:
@@ -572,18 +792,51 @@ conservative_parameter_merge: ParameterMerger
 def conservative_parameter_merge(
     params, same_kind=True, same_default=True, same_annotation=True
 ):
-    """Validates that all the params are exactly the same, returning the first is so."""
+    """Validates that all the params are exactly the same, returning the first if so.
+
+    This is used when hooking up functions that use the same parameters (i.e. arg
+    names). When the name of an argument is used more than once, which kind, default,
+    and annotation should be used in the interface of the DAG?
+
+    If they're all the same, there's no problem.
+
+    But if they're not the same, we need to provide control on which to ignore.
+
+    """
+    suggestion_on_error = '''To resolve this you have several choices:
+    
+    - Change the properties of the param (kind, default, annotation) to be those you 
+      want. For example, you can use ``i2.Sig.ch_param_attrs`` 
+      (or ``i2.Sig.ch_defaults``, ``i2.Sig.ch_kinds``, ``i2.Sig.ch_annotations``)
+      to get a function decorator that will do that for you.
+    - If you're making a DAG, consider specifying a different ``parameter_merge``.
+      For example you can use ``functools.partial`` on 
+      ``i2.dag.conservative_parameter_merge``, fixing ``same_kind``, ``same_default``, 
+      and/or ``same_annotation`` to ``False`` to get a more lenient version of it.
+      
+    See https://github.com/i2mint/meshed/issues/7 (description and comments) for more
+    info.
+    '''
     first_param, *_ = params
     if not all(p.name == first_param.name for p in params):
-        raise ValidationError(f"Some params didn't have the same name: {params}")
+        raise ValidationError(
+            f"Some params didn't have the same name: {params}\n{suggestion_on_error}"
+        )
     if same_kind and not all(p.kind == first_param.kind for p in params):
-        raise ValidationError(f"Some params didn't have the same kind: {params}")
+        raise ValidationError(
+            f"Some params didn't have the same kind: {params}\n{suggestion_on_error}"
+        )
     if same_default and not all(p.default == first_param.default for p in params):
-        raise ValidationError(f"Some params didn't have the same default: {params}")
+        raise ValidationError(
+            f"Some params didn't have the same default: {params}\n{suggestion_on_error}"
+        )
     if same_annotation and not all(
         p.annotation == first_param.annotation for p in params
     ):
-        raise ValidationError(f"Some params didn't have the same annotation: {params}")
+        raise ValidationError(
+            f"Some params didn't have the same annotation: "
+            f'{params}\n{suggestion_on_error}'
+        )
     return first_param
 
 
@@ -614,6 +867,10 @@ def partialized_funcnodes(func_nodes, **keyword_defaults):
             yield func_node
 
 
+Scope = dict
+VarNames = Iterable[str]
+DagOutput = Any
+
 # TODO: caching last scope isn't really the DAG's direct concern -- it's a debugging
 #  concern. Perhaps a more general form would be to define a cache factory defaulting
 #  to a dict, but that could be a "dict" that logs writes (even to an attribute of self)
@@ -671,6 +928,9 @@ class DAG:
     # can return a prepopulated scope too!
     new_scope: Callable = field(default=dict, repr=False)
     name: str = None
+    extract_output_from_scope: Callable[[Scope, VarNames], DagOutput] = field(
+        default=extract_values, repr=False
+    )
 
     def __post_init__(self):
         self.func_nodes = tuple(_mk_func_nodes(self.func_nodes))
@@ -679,7 +939,11 @@ class DAG:
         # reorder the nodes to fit topological order
         self.func_nodes, self.var_nodes = _separate_func_nodes_and_var_nodes(self.nodes)
         # self.sig = Sig(dict(extract_items(sig.parameters, 'xz')))
-        self.sig = Sig(sort_params(self.src_name_params(root_nodes(self.graph))))
+        self.sig = Sig(  # make a signature
+            sort_params(  # with the sorted params (sorted to satisfy kind/default order)
+                self.src_name_params(root_nodes(self.graph))
+            )
+        )
         self.sig(self)  # to put the signature on the callable DAG
         # figure out the roots and leaves
         self.roots = tuple(self.sig.names)  # roots in the same order as signature
@@ -690,20 +954,22 @@ class DAG:
         if self.name is not None:
             self.__name__ = self.name
 
-    def _call(self, *args, **kwargs):
-        scope = self.sig.kwargs_from_args_and_kwargs(args, kwargs)
-        self.call_on_scope(scope)
-        tup = tuple(extract_values(scope, self.leafs))
-        if len(tup) > 1:
-            return tup
-        elif len(tup) == 1:
-            return tup[0]
-        else:
-            return None
-
     def __call__(self, *args, **kwargs):
         return self._call(*args, **kwargs)
 
+    def _call(self, *args, **kwargs):
+        # Get a dict of {argname: argval} pairs from positional and keyword arguments
+        # How positionals are resolved is determined by sels.sig
+        # The result is the initial ``scope`` the func nodes will both read from
+        # to get their arguments, and write their outputs to.
+        scope = self.sig.kwargs_from_args_and_kwargs(args, kwargs)
+        # Go through self.func_nodes in order and call them on scope (performing said
+        # read_input -> call_func -> write_output operations)
+        self.call_on_scope(scope)
+        # From the scope, that may contain all intermediary results,
+        # extract the desired final output and return it
+        return self.extract_output_from_scope(scope, self.leafs)
+
     def call_on_scope(self, scope=None):
         """Calls the func_nodes using scope (a dict or MutableMapping) both to
         source it's arguments and write it's results.
@@ -832,21 +1098,49 @@ class DAG:
 
     # ------------ utils --------------------------------------------------------------
 
-    def src_name_params(self, bind: Optional[Iterable[str]] = None):
+    @property
+    def params_for_src(self):
+        """The ``{src_name: list_of_params_using_that_src,...}`` dictionary.
+        That is, a ``dict`` having lists of all ``Parameter`` objs that are used by a
+        ``node.bind`` source (value of ``node.bind``) for each such source in the graph
+
+        For each ``func_node``, ``func_node.bind`` gives us the
+        ``{param: varnode_src_name}`` specification that tells us where (key of scope)
+        to source the arguments of the ``func_node.func`` for each ``param`` of that
+        function.
+
+        What ``params_for_src`` is, is the corresponding inverse map.
+        The ``{varnode_src_name: list_of_params}`` gathered by scanning each
+        ``func_node`` of the DAG.
+        """
         d = defaultdict(list)
         for node in self.func_nodes:
             for arg_name, src_name in node.bind.items():
                 d[src_name].append(node.sig.parameters[arg_name])
+        return dict(d)
 
-        if bind is None:
-            bind = set(d)
+    def src_name_params(self, src_names: Optional[Iterable[str]] = None):
+        # see params_for_src property to see what d is
+        d = self.params_for_src
+        if src_names is None:  # if no src_names given, use the names of all var_nodes
+            src_names = set(d)
 
-        for src_name in filter(bind.__contains__, d):
-            params = d[src_name]
-            if len(params) == 1:
-                yield params[0].replace(name=src_name)
-            else:
-                yield self.parameter_merge(params).replace(name=src_name)
+        # For every src_name of the DAG that is in ``src_name``...
+        for src_name in filter(src_names.__contains__, d):
+            params = d[src_name]  # consider all the params that use it
+            # make version of these params that have the same name (namely src_name)
+            params_with_name_changed_to_src_name = [
+                p.replace(name=src_name) for p in params
+            ]
+            if len(params_with_name_changed_to_src_name) == 1:
+                # if there's only one param, return it (there can be no conflict)
+                yield params_with_name_changed_to_src_name[0]
+            else:  # if there's more than one param, merge them
+                # How to resolve conflicts (different defaults, annotations or kinds)
+                # is determined by what ``parameter_merge`` specified, which is,
+                # by default, strict (everything needs to be the same, or
+                # ``parameter_merge`` with raise an error.
+                yield self.parameter_merge(params_with_name_changed_to_src_name)
 
     # ------------ display --------------------------------------------------------------
 
@@ -873,8 +1167,8 @@ class DAG:
                 f'{e}\nYou may not have graphviz installed. '
                 f'See https://pypi.org/project/graphviz/.'
             )
-
-        body = list(self.dot_digraph_body(*args, **kwargs))
+        # Note: Since graphviz 0.18, need to have a newline in body lines!
+        body = list(map(_add_new_line_if_none, self.dot_digraph_body(*args, **kwargs)))
         return graphviz.Digraph(body=body)
 
 
@@ -1042,6 +1336,15 @@ def dot_lines_of_func_node(func_node: FuncNode):
     )
 
 
+def _add_new_line_if_none(s: str):
+    """Since graphviz 0.18, need to have a newline in body lines.
+    This util is there to address that, adding newlines to body lines
+    when missing."""
+    if s and s[-1] != '\n':
+        return s + '\n'
+    return s
+
+
 # ---------- with ext.gk -------------------------------------------------------
 
 with suppress(ModuleNotFoundError, ImportError):
@@ -1064,3 +1367,14 @@ with suppress(ModuleNotFoundError, ImportError):
 
         for func, operation in zip(funcs, funcs_to_operations(funcs, exclude_names)):
             yield operation(func)
+
+
+if __name__ == '__main__':
+    from meshed import FuncNode
+    from lined import Line
+
+    def foo(a=1):
+        return a + 1
+
+    line = Line(foo)
+    fn_line = FuncNode(line)
